import json
import re
from copy import deepcopy

import pyarrow.parquet as pq
from fsspec.implementations.local import LocalFileSystem
from fsspec.spec import AbstractFileSystem
from pyarrow.dataset import dataset as pyarrow_dataset


class DeltaTable:
    """
    Access Delta tables on different filesystems.

    Parameters
    -----------
    path:  string
        Path to the table in the file system.
        Note that for Azure, AWS and GCP you need to include the container/bucket name in the path
    file_system: FSSpec compliant filesystem
        Filesystem to be used for reading the Delta files.

    """

    def __init__(self, path, file_system=None):
        if not isinstance(file_system, AbstractFileSystem) and file_system is not None:
            raise TypeError("file_system must be a fsspec compliant filesystem")

        self.path = path
        self.log_path = f"{self.path}/_delta_log"
        self.version = 0
        self.checkpoint = 0
        self.files = set()
        if file_system is None:
            file_system = LocalFileSystem()
        self.filesystem = file_system
        self._as_newest_version()

        # The PyArrow Dataset is exposed by a factory class,
        # which makes it hard to inherit from it directly.
        # Instead we will just have the dataset as an attribute and expose the important methods.
        self.pyarrow_dataset = pyarrow_dataset(
            source=list(self.files),
            filesystem=self.filesystem,
            partitioning="hive",
            format="parquet",
        )

    @property
    def schema(self):
        return self.pyarrow_dataset.schema

    def _is_delta_table(self):
        return self.filesystem.exists(f"{self.log_path}/{0:020}.json")

    def _apply_from_checkpoint(self, checkpoint_version: int):

        # reset file set, and checkpoint version
        self.files = set()
        self.checkpoint = checkpoint_version

        if self.checkpoint == 0:
            return

        # read latest checkpoint
        with self.filesystem.open(
            f"{self.log_path}/{self.checkpoint:020}.checkpoint.parquet"
        ) as checkpoint_file:
            checkpoint = pq.read_table(checkpoint_file).to_pandas()

            for i, row in checkpoint.iterrows():
                added_file = row["add"]["path"] if row["add"] else None
                if added_file:
                    self.files.add(f"{self.path}/{added_file}")

    def _apply_partial_logs(self, version: int):
        # Checkpoints are created every 10 transactions,
        # so we need to find all log files with version
        # up to 9 higher than checkpoint.
        # Effectively, this means that we can just create a
        # wild card for the first decimal of the checkpoint version

        log_files = self.filesystem.glob(
            f"{self.log_path}/{self.checkpoint//10:019}*.json"
        )
        # sort the log files, so we are sure we get the correct order
        log_files = sorted(log_files)
        for log_file in log_files:

            # Get version from log name
            log_version = re.findall(r"(\d{20})", log_file)[0]
            self.version = int(log_version)

            # Download log file
            log = self.filesystem.cat(log_file)
            for line in log.split():
                meta_data = json.loads(line)
                # Log contains other stuff, but we are only
                # interested in the add or remove entries
                if "add" in meta_data.keys():
                    self.files.add(f"{self.path}/{meta_data['add']['path']}")
                if "remove" in meta_data.keys():
                    remove_file = meta_data["remove"]["path"]
                    # To handle 0 checkpoints, we might read the log file with
                    # same version as checkpoint. this means that we try to
                    # remove a file that belongs to an ealier version,
                    # which we don't have in the list
                    if remove_file in self.files:
                        self.files.remove(f"{self.path}/{remove_file}")
            # Stop if we have reatched the desired version
            if self.version == version:
                break

    def _as_newest_version(self):
        # Try to get the latest checkpoint info
        try:
            # get latest checkpoint version
            checkpoint_info = self.filesystem.cat(f"{self.log_path}/_last_checkpoint")
            checkpoint_info = json.loads(checkpoint_info)
            self._apply_from_checkpoint(checkpoint_info["version"])

        except FileNotFoundError:
            pass

        # apply remaining versions. This can be a maximum of 9 versions.
        # we will just break when we don't find any newer logs
        self._apply_partial_logs(version=self.checkpoint + 9)

    def to_table(self, *args, **kwargs):
        """
        Convert to a pyarrow Table.
        Is based on the `to_pandas` function from `pyarrow.Table.to_pandas`,
        so any this will accept the same arguments.
        For more information see https://arrow.apache.org/docs/python/generated/pyarrow.dataset.FileSystemDataset.html#pyarrow.dataset.FileSystemDataset.to_table
        """  # noqa E501
        return self.pyarrow_dataset.to_table(*args, **kwargs)

    def to_pandas(self, *args, **kwargs):
        """
        Convert to a pandas dataframe.
        Is based on the `to_pandas` function from `pyarrow.Table.to_pandas`,
        so any this will accept the same arguments.
        For more information see https://arrow.apache.org/docs/python/generated/pyarrow.Table.html?highlight=to_pandas#pyarrow.Table.to_pandas
        """  # noqa E501
        return self.to_table().to_pandas(*args, **kwargs)

    def as_version(self, version: int, inplace=True):
        """
        Find the files for a specific version of the table.

        Parameters:
        ----------
        version: int
            The table version number that should be loaded

        inplace: Bool
            Specify wether the object should be modified inplace or not.
            If `True`, the current object will be modified.
            if `False`, a new instance of the `DeltaTable` will be returned with the given version.

        Returns:
        -------
        dr : (DeltaTable)
            Delta table that has parsed the log files for the specific version
        """
        nearest_checkpoint = version // 10
        if inplace:
            self._apply_from_checkpoint(nearest_checkpoint)
            self._apply_partial_logs(version=version)
            self.pyarrow_dataset = pyarrow_dataset(
                source=list(self.files),
                filesystem=self.filesystem,
                partitioning="hive",
                format="parquet",
            )
            return self

        deltaTable = deepcopy(self)
        deltaTable._apply_from_checkpoint(nearest_checkpoint)
        deltaTable._apply_partial_logs(version=version)
        deltaTable.pyarrow_dataset = pyarrow_dataset(
            source=list(deltaTable.files),
            filesystem=deltaTable.filesystem,
            partitioning="hive",
            format="parquet",
        )

        return deltaTable



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['deltalake']

package_data = \
{'': ['*']}

install_requires = \
['fsspec>=0.8.5,<0.9.0',
 'pandas>=1.1.5,<2.0.0',
 'pip>=20.3.3,<21.0.0',
 'pyarrow>=2.0.0,<3.0.0']

extras_require = \
{'aws': ['s3fs>=0.5.2,<0.6.0'],
 'azure': ['adlfs>=0.5.9,<0.6.0', 'aiohttp>=3.7.3,<4.0.0'],
 'gcp': ['gcsfs>=0.7.1,<0.8.0']}

setup_kwargs = {
    'name': 'delta-lake-reader',
    'version': '0.2.1',
    'description': 'Lightweight wrapper for reading Delta tables without Spark',
    'long_description': '![Build package](https://github.com/jeppe742/DeltaLakeReader/workflows/Build%20python%20package/badge.svg)\n# Delta Lake Reader\nThe [Delta](https://github.com/delta-io/delta) format, developed by Databricks, is often used to build data lakes.\n\nWhile it tries to solve many issues with data lakes, one of the downsides is that delta tables rely on Spark to read the data. If you only need to read a small table, this can introduce a lot of unnecessary overhead.\n\nThis package tries to fix this, by providing a lightweight python wrapper around the delta file format, **without** any Spark dependencies\n\n\n# Installation\nInstall the package using pip\n\n\n```\npip install delta-lake-reader\n```\nThis will only install the minimal dependencies for working with local file system.\nTo access Delta Tables stored in popular cloud storages, use one of the following commands, to include the cloud specific dependencies\n\n### Azure\n```\npip install delta-lake-reader[azure]\n```\n### Amazon Web Services (AWS)\n```\npip install delta-lake-reader[aws]\n```\n### Google Cloud Platform (GCP)\n```\npip install delta-lake-reader[gcp]\n```\n# Usage\nPackage is build on [PyArrow](https://arrow.apache.org/docs/python/) and [FSSpec](https://filesystem-spec.readthedocs.io/en/latest/index.html).\n\nThis means that you get all the features of PyArrow, like predicate pushdown, partition pruning and easy interoperability with Pandas.\n\nMeanwhile, FSSpec serves as a FileSystem agnostic backend, that lets you read files from many places, including popular cloud providers.\n\nTo read a DeltaTable, first create a `DeltaTable` object.\nThis will read the delta transaction log to find the current files, and get the schema. This will, however, **not** read any data.\nTo read the content of the table, call `to_table()` to get a `pyarrow.Table` object, or `to_pandas()` to get a `pandas.DataFrame`\n## Local file system\n\n```python\nfrom deltalake import DeltaTable\n\n# native file path. Can be relative or absolute\ntable_path = "somepath/mytable"\n\n# Get table as pyarrow table\ndf = DeltaTable(table_path).to_table()\n\n# Get table as pandas dataframe\ndf = DeltaTable(table_path).to_pandas()\n```\n## Azure\nThe Azure integration is based on [adlfs](https://github.com/dask/adlfs) package, developed by the [Dask](https://dask.org/) community.\n\n\n\n`credential` used to authenticate against the storage account, can be either a SAS token, Access Keys or one of the `azure.identity` classes. See [authentication using Azure SDK](https://docs.microsoft.com/en-us/python/api/azure-storage-blob/azure.storage.blob.aio.blobserviceclient?view=azure-python#parameters) for more information.\n\n```python\nfrom deltalake import DeltaTable\nfrom adlfs import AzureBlobFileSystem\n\n#example url  \'abfss://myContainer@myStorageAccount.dfs.core.windows.net/somepath/mytable\'\nfs = AzureBlobFileSystem(\n        account_name="myStorageAccount", \n        credential=\'...\'\n    )\ndf = DeltaTable("mycontainer/somepath/mytable", file_system=fs).to_pandas()\n\n```\n\n## Amazon Web Service (AWS)\nThe AWS integration is based on [s3fs](https://s3fs.readthedocs.io/en/latest/) package, developed by the [Dask](https://dask.org/) community. \n\nTo authenticate you can either specify the access key and secret, or since it\'s build on `boto`, use one of their methods for authentication. See [authentication using AWS SDK ](https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html) for more information.\n\n```python\nfrom deltalake import DeltaTable\nfrom s3fs import S3FileSystem\n\n#example url  \'s3://myBucket/somepath/mytable\'\nfs = S3FileSystem() #authenticate using environment variables, in this example\ndf = DeltaTable("myBucket/somepath/mytable", file_system=fs).to_pandas()\n\n```\n\n## Google Cloud Platform (GCP)\nThe GCP integration is based on the [gcsfs](https://gcsfs.readthedocs.io/en/latest/), developed by the [Dask](https://dask.org/) community.\n\nFor more information about authentication with GCP see the [gcsfs documentation](https://gcsfs.readthedocs.io/en/latest/#credentials) or the [GCP documentation](https://cloud.google.com/docs/authentication)\n\n```python\nfrom deltalake import DeltaTable\nfrom gcsfs import GCSFileSystem\n\n#example url  \'gs://myBucket/somepath/mytable\'\nfs = GCSFileSystem() #authenticate using environment variables, in this example\ndf = DeltaTable("myBucket/somepath/mytable", file_system=fs).to_pandas()\n\n```\n\n## Time travel\nOne of the features of the Delta format, is the ability to do [timetravel](https://docs.delta.io/0.3.0/delta-batch.html#query-an-older-snapshot-of-a-table-time-travel).\n\nThis can be done using the `as_version` method. Note that this currenly only support specific version, and not timestamp.\n```python\nfrom deltalake import DeltaTable\n\ndf = DeltaTable("somepath/mytable").as_version(5).to_pandas()\n```\nTimetraveling to a version that has been vacuumed, currently results in undefined behavior\n\n\n## Predicate Pushdown, Partition Pruning & Columnar file formats\nSince the resulting `DeltaTable` is based on the `pyarrow.DataSet`, you get many cool features for free. \n\nThe `DeltaTable.to_table` is inherited from `pyarrow.Dataset.to_table`. This means that you can include arguments like `filter`, which will do partition pruning and predicate pushdown. If you have a partitioned dataset, partition pruning can potentially reduce the data needed to be downloaded substantially. The predicate pushdown will not have any effect on the amount of data downloaded, but will reduce the dataset size when loaded into memory.\n\nFurther more, since the underlying parquet file format is columnar, you can select a subset of columns to be read from the files. This can be done by passing a list of column names to `to_table`.\n\nSee documentation of [to_pandas](https://arrow.apache.org/docs/python/generated/pyarrow.Table.html#pyarrow.Table.to_pandas), or [to_table](https://arrow.apache.org/docs/python/generated/pyarrow.dataset.FileSystemDataset.html#pyarrow.dataset.FileSystemDataset.to_table) for documentation of all arguments\n\n```python\nimport pyarrow.dataset as ds\n\n#Predicate pushdown. \n#If the table is partitioned on age, it will also to partition pruning\ndf = DeltaTable("...").to_table(filter=ds.field("age")>=18).to_pandas()\n\n#Only load a subset of columns\ndf = DeltaTable("...").to_table(columns=["age","name"]).to_pandas()\n\n```\n\n[Read more about filtering data using PyArrow](https://arrow.apache.org/docs/python/dataset.html#filtering-data)\n\n\n\n\n## Bring Your Own Filesystem\nSince the implementation is using the FSSpec for filesystem abstraction, you can in principle use any FSSpec filesystem. [See more about available FSSpec interfaces](https://filesystem-spec.readthedocs.io/en/latest/api.html#built-in-implementations).\n\n```python\nfs = SomeFSSpecFilesystem()\ndf = DeltaTable(path, file_system=fs).to_pandas()\n```\n\n## Disclaimer\nDatabricks recently announced a stand alone reader for Delta tables in a [blogpost](https://databricks.com/blog/2020/12/22/natively-query-your-delta-lake-with-scala-java-and-python.html)\nThe stand alone reader is JVM based, but a [Rust library](https://github.com/delta-io/delta-rs) with python bindings is also mentioned. This, however, cannot be pip installed which may discourage many python developers. \nAlthough the idea for this library was made independently, some inspirations has been taken from the Rust library.\n# Read more\n[Delta transaction log](https://databricks.com/blog/2019/08/21/diving-into-delta-lake-unpacking-the-transaction-log.html)\n\n[PyArrow Documentation](https://arrow.apache.org/docs/index.html)\n\n[FSSpec Documentation](https://filesystem-spec.readthedocs.io/en/latest/)\n',
    'author': 'Jeppe Johan Waarkjær Olsen',
    'author_email': 'jeppe742@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jeppe742/DeltaLakeReader',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
